#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csiactobs script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# =============================== #
# Test class for csiactobs script #
# =============================== #
class Test(test):
    """
    Test class for csiactobs script

    This test class makes unit tests for the csiactobs script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set data members
        self._datapath = self._datadir + '/iactdata'
        self._runlist  = self._datadir + '/iact_runlist.dat'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('csiactobs')

        # Append tests
        self.append(self._test_cmd, 'Test csiactobs on command line')
        self.append(self._test_python, 'Test csiactobs from Python')

        # Return
        return

    # Test csiactobs on command line
    def _test_cmd(self):
        """
        Test csiactobs on the command line
        """
        # Set script name
        csiactobs = self._script('csiactobs')

        # Setup csiactobs command
        cmd = csiactobs+' datapath="'+self._datapath+'"'+ \
                        ' prodname="unit-test"'+ \
                        ' infile="'+self._runlist+'"'+ \
                        ' bkgpars=1'+\
                        ' outobs="csiactobs_obs_cmd1.xml"'+ \
                        ' outmodel="csiactobs_bgd_cmd1.xml"'+ \
                        ' logfile="csiactobs_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check observation definition XML file
        self._check_obsdef('csiactobs_obs_cmd1.xml', 6)

        # Check model definition XML file
        self._check_moddef('csiactobs_bgd_cmd1.xml', 6)

        # Setup csiactobs command
        cmd = csiactobs+' datapath="data_path_that_does_not_exist"'+ \
                        ' prodname="unit-test"'+ \
                        ' infile="'+self._runlist+'"'+ \
                        ' bkgpars=1'+\
                        ' outobs="csiactobs_obs_cmd2.xml"'+ \
                        ' outmodel="csiactobs_bgd_cmd2.xml"'+ \
                        ' logfile="csiactobs_cmd2.log" debug=yes debug=yes'+ \
                        ' chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input datapath when executed from command line')
        
        # Setup csiactobs command
        cmd = csiactobs+' datapath="'+self._datapath+'"'+ \
                        ' prodname="unit-test-doesnt-exist"'+ \
                        ' infile="'+self._runlist+'"'+ \
                        ' bkgpars=1'+\
                        ' outobs="csiactobs_obs_cmd3.xml"'+ \
                        ' outmodel="csiactobs_bgd_cmd3.xml"'+ \
                        ' logfile="csiactobs_cmd3.log" debug=yes debug=yes'+ \
                        ' chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input prodname when executed from command line')

        # Check csiactobs --help
        self._check_help(csiactobs)

        # Return
        return

    # Test csiactobs from Python
    def _test_python(self):
        """
        Test csiactobs from Python
        """
        # Allocate empty csiactobs script
        iactobs = cscripts.csiactobs()

        # Check that empty csiactobs sciript has an empty observation container
        # and energy boundaries
        self.test_value(iactobs.obs().size(), 0,
             'Check that empty csiactobs has an empty observation container')
        self.test_value(iactobs.ebounds().size(), 0,
             'Check that empty csiactobs has empty energy bins')

        # Check that saving saves an empty model definition file
        iactobs['outobs']   = 'csiactobs_obs_py0.xml'
        iactobs['outmodel'] = 'csiactobs_bgd_py0.xml'
        iactobs['logfile']  = 'csiactobs_py0.log'
        iactobs.logFileOpen()
        iactobs.save()

        # Check empty observation definition XML file
        self._check_obsdef('csiactobs_obs_py0.xml', 0)

        # Check empty model definition XML file
        self._check_moddef('csiactobs_bgd_py0.xml', 0)

        # Check that clearing does not lead to an exception or segfault
        #iactobs.clear()

        # Set-up csiactobs
        iactobs = cscripts.csiactobs()
        iactobs['datapath'] = self._datapath
        iactobs['prodname'] = 'unit-test'
        iactobs['infile']   = self._runlist
        iactobs['bkgpars']  = 1
        iactobs['outobs']   = 'csiactobs_obs_py1.xml'
        iactobs['outmodel'] = 'csiactobs_bgd_py1.xml'
        iactobs['logfile']  = 'csiactobs_py1.log'
        iactobs['chatter']  = 2

        # Run csiactobs script and save run list
        iactobs.logFileOpen()   # Make sure we get a log file
        iactobs.run()
        iactobs.save()

        # Check observation definition XML file
        self._check_obsdef('csiactobs_obs_py1.xml', 6)
        
        # Check model definition XML file
        self._check_moddef('csiactobs_bgd_py1.xml', 6)
        
        # Create test runlist
        runlist = ['15000','15001']
           
        # Set-up csiactobs using a runlist with 2 background parameters
        iactobs = cscripts.csiactobs()
        iactobs['datapath'] = self._datapath
        iactobs['prodname'] = 'unit-test'
        iactobs['bkgpars']  = 2
        iactobs['outobs']   = 'csiactobs_obs_py2.xml'
        iactobs['outmodel'] = 'csiactobs_bgd_py2.xml'
        iactobs['logfile']  = 'csiactobs_py2.log'
        iactobs['chatter']  = 3
        iactobs.runlist(runlist)
   
        # Run csiactobs script and save run list
        iactobs.logFileOpen()   # Make sure we get a log file
        iactobs.run()
        iactobs.save()
           
        # Test return functions
        self.test_value(iactobs.obs().size(), 2,
                        'Check number of observations in container')
        self.test_value(iactobs.ebounds().size(), 0,
                        'Check number of energy boundaries')
   
        # Check observation definition XML file
        self._check_obsdef('csiactobs_obs_py2.xml',2)
           
        # Check model definition XML file
        self._check_moddef('csiactobs_bgd_py2.xml',2)
           
        # Set-up csiactobs with a large number of free parameters and "aeff"
        # background
        iactobs = cscripts.csiactobs()
        iactobs['datapath']      = self._datapath
        iactobs['prodname']      = 'unit-test'
        iactobs['infile']        = self._runlist
        iactobs['bkgpars']       = 8
        iactobs['bkg_mod_hiera'] = 'aeff'
        iactobs['outobs']        = 'csiactobs_obs_py3.xml'
        iactobs['outmodel']      = 'csiactobs_bgd_py3.xml'
        iactobs['logfile']       = 'csiactobs_py3.log'
        iactobs['chatter']       = 4
    
        # Execute csiactobs script
        iactobs.execute()
        
        # Check observation definition XML file
        self._check_obsdef('csiactobs_obs_py3.xml',6)
            
        # Check model definition XML file
        self._check_moddef('csiactobs_bgd_py3.xml',6)
            
        # Set-up csiactobs with a "gauss" background and "inmodel" parameter
        iactobs = cscripts.csiactobs()
        iactobs['datapath']      = self._datapath
        iactobs['inmodel']       = self._model
        iactobs['prodname']      = 'unit-test'
        iactobs['infile']        = self._runlist
        iactobs['bkgpars']       = 1
        iactobs['bkg_mod_hiera'] = 'gauss'
        iactobs['outobs']        = 'NONE'
        iactobs['outmodel']      = 'NONE'
        iactobs['logfile']       = 'csiactobs_py4.log'
        iactobs['chatter']       = 4
   
        # Run csiactobs script
        iactobs.logFileOpen()   # Make sure we get a log file
        iactobs.run()
        
        # Check number of observations
        self.test_value(iactobs.obs().size(), 6,
                        'Check number of observations in container')

        # Check number of models
        self.test_value(iactobs.obs().models().size(), 8,
                        'Check number of models in container')        
        
        # Set-up csiactobs with a "gauss" background and "inmodel" parameter
        iactobs = cscripts.csiactobs()
        iactobs['datapath']      = self._datapath
        iactobs['inmodel']       = self._model
        iactobs['prodname']      = 'unit-test'
        iactobs['infile']        = self._runlist
        iactobs['bkgpars']       = 1
        iactobs['bkg_mod_hiera'] = 'irf'
        iactobs['outobs']        = 'NONE'
        iactobs['outmodel']      = 'NONE'
        iactobs['logfile']       = 'csiactobs_py4.log'
        iactobs['chatter']       = 4
   
        # Run csiactobs script
        iactobs.logFileOpen()   # Make sure we get a log file
        iactobs.run()
        
        # Check number of observations
        self.test_value(iactobs.obs().size(), 5,
                        'Check number of observations in container')

        # Check number of models
        self.test_value(iactobs.obs().models().size(), 7,
                        'Check number of models in container')     

        # Return
        return

    # Check observation definition XML file
    def _check_obsdef(self, filename, obs_expected):
        """
        Check observation definition XML file
        """
        # Load observation definition XML file
        obs = gammalib.GObservations(filename)

        # Check number of observations
        self.test_value(obs.size(), obs_expected,
                        'Check for '+str(obs_expected)+' observations in XML file')
        
        # If there are observations in the XML file then check their content
        if obs_expected > 0:
            
            # Get response
            rsp = obs[0].response()
            
            # Test response
            self.test_value(obs[0].eventfile().file(), 'events_0.fits.gz',
                            'Check event file name')
            self.test_value(obs[0].eventfile().extname(), 'EVENTS',
                            'Check event extension name')
            self.test_value(rsp.aeff().filename().file(), 'irf_file.fits.gz',
                            'Check effective area file name')
            self.test_value(rsp.aeff().filename().extname(), 'EFFECTIVE AREA',
                            'Check effective area extension name')
            self.test_value(rsp.psf().filename().file(), 'irf_file.fits.gz',
                            'Check point spread function file name')
            self.test_value(rsp.psf().filename().extname(), 'POINT SPREAD FUNCTION',
                            'Check point spread function extension name')
            self.test_value(rsp.edisp().filename().file(), 'irf_file.fits.gz',
                            'Check energy dispersion file name')
            self.test_value(rsp.edisp().filename().extname(), 'ENERGY DISPERSION',
                            'Check energy dispersion extension name')
            self.test_value(rsp.background().filename().file(), 'irf_file.fits.gz',
                            'Check background file name')
            self.test_value(rsp.background().filename().extname(), 'BACKGROUND',
                            'Check background extension name')
        
        # Return
        return

    # Check model XML file
    def _check_moddef(self, filename, models_expected):
        """
        Check model definition XML file
        """
        # Load model definition XML file
        models = gammalib.GModels(filename)

        # Check number of models
        self.test_value(models.size(), models_expected,
                        'Check for '+str(models_expected)+' models in XML file')

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the mputils module
#
# Copyright (C) 2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import ctools
from cscripts import mputils
from testing import test


# ================== #
# mputils test class #
# ================== #
class mputils_test(ctools.cscript):
    """
    Test class for the mputils module
    """
    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the base class constructor
        self._init_cscript(self.__class__.__name__, ctools.__version__, argv)

        # Return
        return

    # Bin computation
    def _compute_bin(self, i):
        """
        Bin computation
        """
        # Return bin index
        return i


# ============================= #
# Test class for mputils module #
# ============================= #
class Test(test):
    """
    Test class for mputils module
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Create parameter file for mputils_test application
        pars = gammalib.GApplicationPars()
        pars.append_standard()
        pars.save('mputils_test.par')

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('mputils')

        # Append tests
        self.append(self._test_nthreads, 'Test mputils.nthreads() function')
        self.append(self._test_process, 'Test mputils.process() function')
        self.append(self._test_mpfunc, 'Test mputils.mpfunc() function')

        # Return
        return

    # Setup mputils test class
    def _setup_mputils_test(self):
        """
        Setup mputils test class
        """
        # Get test class
        cls = mputils_test()

        # Recover User parameters
        pars = cls.pars()

        # Create a copy of the User parameters
        cpy_pars = pars.copy()

        # Return
        return (cls, pars, cpy_pars)

    # Test nthreads() function
    def _test_nthreads(self):
        """
        Test nthreads() function
        """
        # Set reference number of CPU's
        n0   = 1
        n739 = 1
        try:
            from multiprocessing import cpu_count
            n0   = cpu_count()
            n739 = 739
        except:
            pass

        # Check nthreads=0. In this case the nthreads() function should return
        # the number of CPU's that are available, provided that the
        # multiprocessing module is available. If the module is not available
        # the function returns 1.
        cls, pars, cpy_pars = self._setup_mputils_test()
        pars.append(gammalib.GApplicationPar('nthreads','i','h','0','','',''))
        nthreads = mputils.nthreads(cls)
        self.test_value(nthreads, n0, 'Check nthreads=0')
        cls.pars(cpy_pars)

        # Check nthreads=739. In this case the nthreads() function should
        # return 739, provided that the multiprocessing module is available.
        # If the module is not available the function returns 1.
        cls, pars, cpy_pars = self._setup_mputils_test()
        pars.append(gammalib.GApplicationPar('nthreads','i','h','739','','',''))
        nthreads = mputils.nthreads(cls)
        self.test_value(nthreads, n739, 'Check nthreads=739')
        cls.pars(cpy_pars)

        # Return
        return

    # Test process() function
    def _test_process(self):
        """
        Test process() function
        """
        # Test only if Pool module is available
        try:
            # Import Pool module
            from multiprocessing import Pool

            # Get test class
            cls = mputils_test()

            # Setup arguments for mpfunc() computation
            args = [(cls, '_compute_bin', i) for i in range(10)]

            # Multiprocessing
            poolresults = mputils.process(2, mputils.mpfunc, args)

            # Loop over arguments
            for i in range(10):
                result = poolresults[i][0]
                self.test_value(result, i, 'Check process for argument %d' % i)

        # ... otherwise if Pool is not available then skip this test
        except:
            pass

        # Return
        return

    # Test mpfunc() function
    def _test_mpfunc(self):
        """
        Test mpfunc() function
        """
        # Get test class
        cls = mputils_test()

        # Setup arguments for mpfunc() computation
        args = [(cls, '_compute_bin', i) for i in range(10)]

        # Loop over arguments
        for arg in args:
            result, _ = mputils.mpfunc(arg)
            self.test_value(result, arg[2], 'Check mpfunc for argument %d' % arg[2])

        # Return
        return



# ==========================================================================
# Utility functions for input and output
#
# Copyright (C) 2016-2017 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import csv
import sys
import gammalib


# ========================= #
# Write one row of CSV file #
# ========================= #
def write_csv_row(outfile, row, colnames, colvalues):
    """
    Write one row into CSV file

    Parameters
    ----------
    outfile : str
        Output file name
    row : int
        Row number
    colnames : list of str
        Column names
    colvalues : list of float
        Column values
    """
    # If this is the first row then open a new file and write the header
    if row == 0:
        f       = open(outfile, 'w')
        writer  = csv.DictWriter(f, colnames)
        headers = {}
        for colname in colnames:
            headers[colname] = colname
        writer.writerow(headers)

    # ... otherwise append to an existing file
    else:
        f = open(outfile, 'a')

    # Write out row
    writer = csv.DictWriter(f, colnames)
    writer.writerow(colvalues)

    # Close file
    f.close()

    # Return
    return


# ================ #
# Read pull values #
# ================ #
def read_pull_values(filename, parname):
    """
    Read pull values from pull distribution ASCII file
    
    Parameters
    ----------
    filename : str
        Pull distribution ASCII file
    parname : str
        Parameter

    Returns
    -------
    values : list of float
        List of pull values

    Raises
    ------
    NameError
    """
    # Initialise list of values
    values = []

    # Open reader
    reader = csv.reader(open(filename, 'r'), delimiter=',')

    # Read rows
    first = True
    index = -1
    for row in reader:

        # Get column index if first row
        if first:
            try:
                index = row.index(parname)
            except:
                print('ERROR: Parameter "'+parname+'" not found in list:')
                for p in row:
                    print('       "'+p+'"')
                raise NameError(parname)

        # Handle data rows
        else:
            values.append(float(row[index]))

        # Flag that first row has been passed
        first = False

    # Return values
    return values


# ============================================ #
# Get arguments and options from argument list #
# ============================================ #
def get_args_options(options, usage):
    """
    Get arguments and options from argument list

    Parameters
    ----------
    options : list of dict
        List of possible options and default values
    usage : str
        Usage string to be shown in case of a problem

    Returns
    -------
    args, options : tuple of list and list of dict
        Arguments and updated list of options
    """
    # Initialise list of arguments
    args = []

    # If there are no command line arguments then show usage string and
    # exit
    if len(sys.argv) < 1:
        print('Usage: %s' % usage)
        sys.exit()

    # First possible option is element 1
    i = 1

    # Loop over all arguments
    while i < len(sys.argv):

        # Initialise that no option was found
        option_found = False

        # Search for options
        for option in options:

            # Do we have an option
            if sys.argv[i] == option['option']:

                # If an option has been found then check if there is a
                # following parameter and extract that parameter as a
                # string. Quit in case that an exception occurs.
                if len(sys.argv) > i+1:
                    i += 1
                    try:
                        option['value'] = str(sys.argv[i])
                    except:
                        print('Usage: %s' % usage)
                        sys.exit()

                # ... there is no following parameter, hence write out usage
                # and quite
                else:
                    print('Usage: %s' % usage)
                    sys.exit()

                # We can break now as every option should only occur once
                option_found = True
                break

        # If no option was found then add the argument to list of arguments
        if not option_found:
            args.append(sys.argv[i])

        # Next item
        i += 1

    # Return arguments and options
    return args, options


# =========================================== #
# Append model parameter column to FITS table #
# =========================================== #
def append_result_column(table, results, name, unit, key):
    """
    Append result column to FITS table

    Parameters
    ----------
    table : `~gammalib.GFitsTable`
        FITS table
    results : list of dict
        Parameter values
    name : str
        Column name
    unit : str
        Column unit
    key : str
        Result key
    """
    # Determine number of rows in FITS table
    nrows = len(results)

    # Allocate double precision column and set units
    col = gammalib.GFitsTableDoubleCol(name, nrows)
    col.unit(unit)

    # Fill results in columns
    for i, result in enumerate(results):
        col[i] = result[key]

    # Append column to table
    table.append(col)

    # Return
    return


# =========================================== #
# Append model parameter column to FITS table #
# =========================================== #
def append_model_par_column(table, model, results):
    """
    Append model parameter column to FITS table

    Parameters
    ----------
    table : `~gammalib.GFitsTable`
        FITS table
    model : `~gammalib.GModel`
        Model
    results : list of dict
        Parameter values
    """
    # Determine number of rows in FITS table
    nrows = len(results)

    # Loop over all parameters in model
    for par in model:

        # Append parameter column only if model parameter is free
        if par.is_free():

            # Set column names
            name   = par.name()
            e_name = 'e_'+par.name()

            # Allocate double precision columns and set units
            col    = gammalib.GFitsTableDoubleCol(name, nrows)
            e_col  = gammalib.GFitsTableDoubleCol(e_name, nrows)
            col.unit(par.unit())
            e_col.unit(par.unit())

            # Fill results in columns
            for i, result in enumerate(results):
                col[i]   = result['values'][name]
                e_col[i] = result['values'][e_name]

            # Append columns to table
            table.append(col)
            table.append(e_col)

    # Return
    return



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the csphagen script.
#
# Copyright (C) 2017-2020 Luigi Tibaldo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================ #
# Test class for csphagen script #
# ================================ #
class Test(test):
    """
    Test class for csphagen script

    This test class makes unit tests for the csphagen script by using it
    from the command line and from Python.
    """

    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set test datasets and parameters
        self._myevents1      = self._datadir + '/crab_offaxis1.fits'
        self._myevents2      = self._datadir + '/crab_offaxis2.fits'
        self._exclusion      = self._datadir + '/crab_exclusion.fits'
        self._nreg_with_excl = 5
        self._nreg_wo_excl   = 8
        self._nreg_bkg_reg   = 5
        self._nreg_mul       = [self._nreg_with_excl, 6]
        self._nbins          = 10
        self._regfile_src    = self._datadir + '/crab_src_reg.reg'
        self._regfile_bkg    = self._datadir + '/crab_bkg_reg.reg'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions.
        """
        # Set test name
        self.name('csphagen')

        # Append tests
        self.append(self._test_cmd, 'Test csphagen on command line')
        self.append(self._test_python, 'Test csphagen from Python')
        self.append(self._test_pickeling, 'Test csphagen pickeling')

        # Return
        return

    # Test csphagen on command line
    def _test_cmd(self):
        """
        Test csphagen on the command line.
        """
        # Set script name
        csphagen = self._script('csphagen')

        # Setup csphagen command
        cmd = csphagen + ' inobs="' + self._myevents1 + \
                         '" inmodel="NONE"' +\
                         ' caldb="' + self._caldb + '" irf="' + self._irf + \
                         '" ebinalg="LOG" emin=0.1 emax=100. enumbins=' + \
                         str(self._nbins) + \
                         ' etruemin=0.05 etruemax=150 etruebins=5'+ \
                         ' coordsys="CEL" ra=83.633 dec=22.0145' + \
                         ' rad=0.2 stack="no" inexclusion="' + \
                         self._exclusion + \
                         '" bkgmethod="REFLECTED" '+ \
                         'outobs="csphagen_cmd1_obs.xml" prefix="csphagen_cmd1" ' + \
                         'outmodel="csphagen_cmd1_model.xml" ' + \
                         'logfile="csphagen_cmd1.log" chatter=2'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
                         'Check successful execution from command line')

        # Check output files
        self._check_output('csphagen_cmd1', self._nbins, self._nreg_with_excl)
        self._check_outobs('csphagen_cmd1_obs.xml', 1)

        # Setup csphagen command
        cmd = csphagen + ' inobs="events_that_do_not_exist.fits" ' + \
                         'inmodel="NONE"' +\
                         ' caldb="' + self._caldb + '" irf="' + self._irf + \
                         '" ebinalg="LOG" emin=0.1 emax=100. enumbins=' + \
                         str(self._nbins) + \
                         ' etruemin=0.05 etruemax=150 etruebins=5'+ \
                         ' coordsys="CEL" ra=83.633 dec=22.0145' + \
                         ' rad=0.2 stack="no" inexclusion="' + \
                         self._exclusion + \
                         '" bkgmethod="REFLECTED" '+ \
                         'outobs="csphagen_cmd2_obs.xml" prefix="csphagen_cmd2" ' + \
                         'outmodel="csphagen_cmd2_model.xml" ' + \
                         'logfile="csphagen_cmd2.log" debug=yes chatter=1'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check csphagen --help
        self._check_help(csphagen)

        # Return
        return

    def _test_python(self):
        """
        Test csphagen from Python
        """
        # Same test as from command line
        phagen = cscripts.csphagen()
        phagen['inobs']       = self._myevents1
        phagen['inmodel']     = 'NONE'
        phagen['caldb']       = self._caldb
        phagen['irf']         = self._irf
        phagen['ebinalg']     = 'LOG'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['coordsys']    = 'CEL'
        phagen['ra']          = 83.633
        phagen['dec']         = 22.0145
        phagen['rad']         = 0.2
        phagen['stack']       = False
        phagen['inexclusion'] = self._exclusion
        phagen['bkgmethod']   = 'REFLECTED'
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py1_obs.xml'
        phagen['outmodel']    = 'csphagen_py1_model.xml'
        phagen['prefix']      = 'csphagen_py1'
        phagen['logfile']     = 'csphagen_py1.log'
        phagen['chatter']     = 1

        # Execute script
        phagen.execute()

        # Check output
        self._check_output('csphagen_py1', self._nbins, self._nreg_with_excl)
        self._check_outobs('csphagen_py1_obs.xml', 1)

        # Now test without exclusion region
        phagen = cscripts.csphagen()
        phagen['inobs']     = self._myevents1
        phagen['inmodel']   = 'NONE'
        phagen['caldb']     = self._caldb
        phagen['irf']       = self._irf
        phagen['ebinalg']   = 'LOG'
        phagen['emin']      = 0.1
        phagen['emax']      = 100.0
        phagen['enumbins']  = self._nbins
        phagen['coordsys']  = 'CEL'
        phagen['ra']        = 83.633
        phagen['dec']       = 22.0145
        phagen['rad']       = 0.2
        phagen['stack']     = False
        phagen['bkgmethod'] = 'REFLECTED'
        phagen['etruemin']  = 0.05
        phagen['etruemax']  = 150.0
        phagen['etruebins'] = 5
        phagen['outobs']    = 'csphagen_py2_obs.xml'
        phagen['outmodel']  = 'csphagen_py2_model.xml'
        phagen['prefix']    = 'csphagen_py2'
        phagen['logfile']   = 'csphagen_py2.log'
        phagen['chatter']   = 2

        # Execute script
        phagen.execute()

        # Check output
        self._check_output('csphagen_py2', self._nbins, self._nreg_wo_excl)
        self._check_outobs('csphagen_py2_obs.xml', 1)

        # Test with multiple input observations, no stacking
        # Create observation container
        obs = gammalib.GObservations()
        for s, events in enumerate([self._myevents1, self._myevents2]):
            run = gammalib.GCTAObservation(events)
            run.id(str(s + 1))
            run.response(self._irf, gammalib.GCaldb('cta', self._caldb))
            obs.append(run)

        # Setup csphagen
        phagen = cscripts.csphagen(obs)
        phagen['inmodel']     = 'NONE'
        phagen['ebinalg']     = 'POW'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['ebingamma']   = 1.0
        phagen['coordsys']    = 'CEL'
        phagen['ra']          = 83.633
        phagen['dec']         = 22.0145
        phagen['rad']         = 0.2
        phagen['stack']       = False
        phagen['inexclusion'] = self._exclusion
        phagen['bkgmethod']   = 'REFLECTED'
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py3_obs.xml'
        phagen['outmodel']    = 'csphagen_py3_model.xml'
        phagen['prefix']      = 'csphagen_py3'
        phagen['logfile']     = 'csphagen_py3.log'
        phagen['chatter']     = 3

        # Run script
        phagen.execute()

        # Check output
        for s in range(2):
            self._check_output('csphagen_py3_' + str(s + 1), self._nbins,
                               self._nreg_mul[s])
        self._check_outobs('csphagen_py3_obs.xml', 2)

        # Setup csphagen for test with multiple input observations and stacking
        phagen = cscripts.csphagen(obs)
        phagen['inmodel']     = 'NONE'
        phagen['ebinalg']     = 'LOG'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['coordsys']    = 'CEL'
        phagen['ra']          = 83.633
        phagen['dec']         = 22.0145
        phagen['rad']         = 0.2
        phagen['stack']       = True
        phagen['inexclusion'] = self._exclusion
        phagen['bkgmethod']   = 'REFLECTED'
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py4_obs.xml'
        phagen['outmodel']    = 'csphagen_py4_model.xml'
        phagen['prefix']      = 'csphagen_py4'
        phagen['logfile']     = 'csphagen_py4.log'
        phagen['chatter']     = 4

        # Execute script
        phagen.execute()

        # Check output
        for s in range(2):
            self._check_output('csphagen_py4_stacked', self._nbins,
                               0, check_regions=False)
        self._check_outobs('csphagen_py4_obs.xml', 1)

        # Setup csphagen for test with custom On and Off regions provided
        phagen = cscripts.csphagen()
        phagen['inobs']      = self._myevents1
        phagen['inmodel']    = 'NONE'
        phagen['caldb']      = self._caldb
        phagen['irf']        = self._irf
        phagen['ebinalg']    = 'LOG'
        phagen['emin']       = 0.1
        phagen['emax']       = 100.0
        phagen['enumbins']   = self._nbins
        phagen['bkgmethod']  = 'CUSTOM'
        phagen['srcregfile'] = self._regfile_src
        phagen['bkgregfile'] = self._regfile_bkg
        phagen['etruemin']   = 0.05
        phagen['etruemax']   = 150.0
        phagen['etruebins']  = 5
        phagen['stack']      = False
        phagen['outobs']     = 'csphagen_py5_obs.xml'
        phagen['outmodel']   = 'csphagen_py5_model.xml'
        phagen['prefix']     = 'csphagen_py5'
        phagen['logfile']    = 'csphagen_py5.log'
        phagen['chatter']    = 2

        # Execute script
        phagen.execute()

        # Check output
        self._check_output('csphagen_py5', self._nbins, self._nreg_bkg_reg)
        self._check_outobs('csphagen_py5_obs.xml', 1)

        # Append off regions to observation container
        for run in obs:
            run.off_regions(gammalib.GSkyRegions(self._regfile_bkg))

        # Setup csphagen for test with multiple input observations and stacking
        phagen = cscripts.csphagen(obs)
        phagen['inmodel']     = 'NONE'
        phagen['ebinalg']     = 'LOG'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['coordsys']    = 'CEL'
        phagen['stack']       = True
        phagen['inexclusion'] = self._exclusion
        phagen['bkgmethod']   = 'CUSTOM'
        phagen['srcregfile']  = self._regfile_src
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py6_obs.xml'
        phagen['outmodel']    = 'csphagen_py6_model.xml'
        phagen['prefix']      = 'csphagen_py6'
        phagen['logfile']     = 'csphagen_py6.log'
        phagen['chatter']     = 4

        # Execute script
        phagen.execute()

        # Check output
        for s in range(2):
            self._check_output('csphagen_py6_stacked', self._nbins,
                               0, check_regions=False)
        self._check_outobs('csphagen_py6_obs.xml', 1)

        # Same test as from command line, but set exclusion with csphagen.exclusion_map
        phagen = cscripts.csphagen()
        phagen['inobs']       = self._myevents1
        phagen['inmodel']     = 'NONE'
        phagen['caldb']       = self._caldb
        phagen['irf']         = self._irf
        phagen['ebinalg']     = 'LOG'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['coordsys']    = 'CEL'
        phagen['ra']          = 83.633
        phagen['dec']         = 22.0145
        phagen['rad']         = 0.2
        phagen['stack']       = False
        phagen['bkgmethod']   = 'REFLECTED'
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py7_obs.xml'
        phagen['outmodel']    = 'csphagen_py7_model.xml'
        phagen['prefix']      = 'csphagen_py7'
        phagen['logfile']     = 'csphagen_py7.log'
        phagen['chatter']     = 1

        # Set exclusion map
        exclmap = gammalib.GSkyRegionMap(self._exclusion)
        phagen.exclusion_map(exclmap)

        # Execute script
        phagen.execute()

        # Check output
        self._check_output('csphagen_py7', self._nbins, self._nreg_with_excl)
        self._check_outobs('csphagen_py7_obs.xml', 1)

        # Return
        return

    # Test csphagen pickeling
    def _test_pickeling(self):
        """
        Test csphagen pickeling
        """
        # Perform pickeling tests of empty class
        self._pickeling(cscripts.csphagen())

        # Set-up csphagen
        phagen = cscripts.csphagen()
        phagen['inobs']       = self._myevents1
        phagen['inmodel']     = 'NONE'
        phagen['caldb']       = self._caldb
        phagen['irf']         = self._irf
        phagen['ebinalg']     = 'LOG'
        phagen['emin']        = 0.1
        phagen['emax']        = 100.0
        phagen['enumbins']    = self._nbins
        phagen['coordsys']    = 'CEL'
        phagen['ra']          = 83.633
        phagen['dec']         = 22.0145
        phagen['rad']         = 0.2
        phagen['stack']       = False
        phagen['inexclusion'] = self._exclusion
        phagen['bkgmethod']   = 'REFLECTED'
        phagen['etruemin']    = 0.05
        phagen['etruemax']    = 150.0
        phagen['etruebins']   = 5
        phagen['outobs']      = 'csphagen_py1_obs_pickle.xml'
        phagen['outmodel']    = 'csphagen_py1_model_pickle.xml'
        phagen['prefix']      = 'csphagen_py1_pickle'
        phagen['logfile']     = 'csphagen_py1_pickle.log'
        phagen['chatter']     = 1

        # Perform pickeling tests of filled class
        obj = self._pickeling(phagen)

        # Run csspec script and save light curve
        obj.logFileOpen()   # Make sure we get a log file
        obj.run()
        obj.save()

        # Check result files
        self._check_output('csphagen_py1_pickle', self._nbins, self._nreg_with_excl)
        self._check_outobs('csphagen_py1_obs_pickle.xml', 1)

        # Return
        return

    def _check_ebounds(self, table, bins):
        """
        Check EBOUNDS table
        """
        # Expected column names
        cols = ['E_MIN', 'E_MAX']

        # Check number of columns, rows and column names
        self.test_value(table.ncols(), len(cols), 'Check for %d columns in '
                        'energy bounds table' % len(cols))
        self.test_value(table.nrows(), bins,
                        'Check for %d rows in energy bounds table' % bins)
        for col in cols:
            self.test_assert(table.contains(col),
                             'Test if EBOUNDS contains "' + col + '" column')

        # Return
        return

    def _check_pha(self, filename, bins, ncols=8):
        """
        Check PHA file
        """
        # Expected column names
        cols = ['CHANNEL', 'COUNTS', 'STAT_ERR', 'SYS_ERR', 'QUALITY',
                'GROUPING', 'AREASCAL', 'BACKSCAL']

        # Open FITS file
        fits = gammalib.GFits(filename)

        # Check FITS file structure
        self.test_value(fits.size(), 3,
                        'Check for 3 extensions in PHA file')
        self.test_assert(fits.contains('SPECTRUM'),
                         'Test if PHA file contains "SPECTRUM" extension')
        self.test_assert(fits.contains('EBOUNDS'),
                         'Test if PHA file contains "EBOUNDS" extension')

        # Get SPECTRUM table
        table = fits['SPECTRUM']

        # Check FITS table structure
        self.test_value(table.ncols(), ncols,
                        'Check for %d columns in PHA table' % len(cols))
        self.test_value(table.nrows(), bins,
                        'Check for %d rows in PHA table' % bins)
        for col in cols:
            self.test_assert(table.contains(col),
                             'Test if PHA file contains "' + col + '" column')

        # Check EBOUNDS table
        table = fits['EBOUNDS']
        self._check_ebounds(table, bins)

        # Close FITS file
        fits.close()

        # Return
        return

    def _check_arf(self, filename, bins):
        """
        Check ARF file
        """
        # Expected column names
        cols = ['ENERG_LO', 'ENERG_HI', 'SPECRESP']

        # Open FITS file
        fits = gammalib.GFits(filename)

        # Check FITS file structure
        self.test_value(fits.size(), 2,
                        'Check for 2 extensions in ARF file')
        self.test_assert(fits.contains('SPECRESP'),
                         'Test if ARF file contains "SPECRESP" extension')

        # Get SPECRESP table
        table = fits['SPECRESP']

        # Check FITS table structure
        self.test_value(table.ncols(), len(cols),
                        'Check for %d columns in ARF table' % len(cols))
        self.test_value(table.nrows(), bins,
                        'Check for %d rows in ARF table' % bins)
        for col in cols:
            self.test_assert(table.contains(col),
                             'Test if ARF file contains "' + col + '" column')

        # Close FITS file
        fits.close()

        # Return
        return

    def _check_rmf(self, filename, bins, etruebins=17):
        """
        Check RMF file
        """
        # Expected column names
        cols = ['ENERG_LO', 'ENERG_HI', 'N_GRP', 'F_CHAN', 'N_CHAN', 'MATRIX']

        # Open FITS file
        fits = gammalib.GFits(filename)

        # Check FITS file structure
        self.test_value(fits.size(), 3,
                        'Check for 3 extensions in RMF file')
        self.test_assert(fits.contains('MATRIX'),
                         'Test if RMF file contains "MATRIX" extension')
        self.test_assert(fits.contains('EBOUNDS'),
                         'Test if RMF file contains "EBOUNDS" extension')

        # Get MATRIX table
        table = fits['MATRIX']

        # Check FITS table structure
        self.test_value(table.ncols(), len(cols),
                        'Check for %d columns in RMF table' % len(cols))
        self.test_value(table.nrows(), etruebins,
                        'Check for %d rows in RMF table' % etruebins)
        for col in cols:
            self.test_assert(table.contains(col),
                             'Test if RMF file contains "' + col + '" column')

        # Check EBOUNDS table
        table = fits['EBOUNDS']
        self._check_ebounds(table, bins)

        # Close FITS file
        fits.close()

        # Return
        return

    def _check_output(self, filenameroot, bins, nreg, check_regions=True):
        """
        Check the output from a csphagen run
        """
        # OGIP files
        self._check_pha(filenameroot + '_pha_on.fits', bins)
        self._check_pha(filenameroot + '_pha_off.fits', bins, ncols=9)
        self._check_arf(filenameroot + '_arf.fits', 17)
        self._check_rmf(filenameroot + '_rmf.fits', bins)

        # Optionally check for regions
        if check_regions:

            # Check On region
            onregion = (filenameroot+'_on.reg').replace('_1_', '_').replace('_2_', '_')
            reg      = gammalib.GSkyRegions(onregion)
            self.test_value(reg.size(), 1, 'Check for 1 region in source region file')

            # Check for Off region
            offregion = filenameroot + '_off.reg'
            reg = gammalib.GSkyRegions(offregion)
            self.test_value(reg.size(), nreg, 'Check for ' + str(nreg) +
                            ' region in background region file')

        # Return
        return

    def _check_outobs(self, filenameroot, nout):
        """
        Check the output XML file containing ON/OFF observations
        """
        # Load observation container
        obs = gammalib.GObservations(filenameroot)

        # Check container size
        self.test_value(obs.size(), nout, 'Check for ' + str(nout) +
                        ' observations in XML file')

        # Return
        return

    def _read_pha_counts(self, filename):
        """
        Read and integrate the counts in a pha file.
        Pha file structure already tested in _check_pha()
        """
        # Open FITS file
        fits = gammalib.GFits(filename)

        # Get SPECTRUM table
        table = fits['SPECTRUM']

        # Integrate counts
        counts_col = table['COUNTS']
        counts     = 0
        for channel in range( counts_col.nrows() ):
            counts += counts_col[channel]

        # Close FITS file
        fits.close()

        # Return
        return counts



#! /usr/bin/env python
# ==========================================================================
# Generates the TS distribution for a particular model
#
# Copyright (C) 2011-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import obsutils
from cscripts import modutils
from cscripts import ioutils
from cscripts import mputils


# ============== #
# cstsdist class #
# ============== #
class cstsdist(ctools.csobservation):
    """
    Generates Test Statistic distribution for a model
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise some members
        self._srcname     = ''
        self._log_clients = False
        self._model       = None
        self._nthreads    = 0

        # Return
        return

    # State methods por pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation getstate method to include some members

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base'        : ctools.csobservation.__getstate__(self),
                 'srcname'     : self._srcname,
                 'log_clients' : self._log_clients,
                 'model'       : self._model,
                 'nthreads'    : self._nthreads}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation setstate method to include some members

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        ctools.csobservation.__setstate__(self, state['base'])
        self._srcname     = state['srcname']
        self._log_clients = state['log_clients']
        self._model       = state['model']
        self._nthreads    = state['nthreads']

        # Return
        return

    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Set observation if not done before
        if self.obs().size() == 0:
            self.obs(self._get_observations())

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Get source name
        self._srcname = self['srcname'].string()

        # Set models if we have none
        if self.obs().models().size() == 0:
            self.obs().models(self['inmodel'].filename())

        # Query parameters
        self['edisp'].boolean()
        self['ntrials'].integer()
        self['debug'].boolean()
        self['outfile'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Return
        return

    def _sim(self, seed):
        """
        Return a simulated observation container

        Parameters
        ----------
        seed : int
            Random number generator seed

        Returns
        -------
        sim : `~gammalib.GObservations`
            Simulated observation container
        """
        # If observation is a counts cube then simulate events from the counts
        # cube model ...
        if self.obs().size() == 1 and self.obs()[0].eventtype() == 'CountsCube':

            # If no counts cube model exists then compute it now
            if self._model == None:
                model            = ctools.ctmodel(self.obs())
                model['debug']   = self['debug'].boolean()
                model['chatter'] = self['chatter'].integer()
                model.run()
                self._model = model.cube().copy() # Save copy for persistence

            # Allocate random number generator
            ran = gammalib.GRan()

            # Get copy of model map
            counts = self._model.counts().copy()

            # Randomize counts
            for i in range(counts.npix()):
                counts[i] = ran.poisson(counts[i])

            # Copy observations
            sim = self.obs().copy()

            # Set counts map
            sim[0].events().counts(counts)

        # ... otherwise simuate events from the observation container (works
        # only for event lists
        else:
            sim = obsutils.sim(self.obs(),
                               seed     = seed,
                               log      = self._log_clients,
                               debug    = self['debug'].boolean(),
                               nthreads = 1)

        # Return simulated observation
        return sim

    def _trial(self, seed):
        """
        Create the TS for a single trial

        Parameters
        ----------
        seed : int
            Random number generator seed

        Returns
        -------
        result : dict
            Result dictionary
        """
        # Write header
        self._log_header2(gammalib.EXPLICIT, 'Trial %d' % (seed+1))

        # Simulate events
        sim = self._sim(seed)

        # Determine number of events in simulation
        nevents = 0.0
        for run in sim:
            nevents += run.events().number()

        # Write simulation results
        self._log_header3(gammalib.EXPLICIT, 'Simulation')
        self._log_value(gammalib.EXPLICIT, 'Number of simulated events', nevents)

        # Fit model
        fit = ctools.ctlike(sim)
        fit['nthreads'] = 1  # Avoids OpenMP conflict
        fit['debug']    = self['debug'].boolean()
        fit['chatter']  = self['chatter'].integer()
        fit.run()

        # Get model fitting results
        logL   = fit.opt().value()
        npred  = fit.obs().npred()
        models = fit.obs().models()
        model  = models[self._srcname]
        ts     = model.ts()

        # Write fit results, either explicit or normal
        if self._logExplicit():
            self._log_header3(gammalib.EXPLICIT, 'Test source model fit')
            self._log_value(gammalib.EXPLICIT, 'Test statistics', ts)
            self._log_value(gammalib.EXPLICIT, 'log likelihood', logL)
            self._log_value(gammalib.EXPLICIT, 'Number of predicted events', npred)
            for model in models:
                self._log_value(gammalib.EXPLICIT, 'Model', model.name())
                for par in model:
                    self._log_string(gammalib.EXPLICIT, str(par))
        elif self._logNormal():
            prefactor = modutils.normalisation_parameter(model)
            name      = 'Trial %d' % seed
            value     = 'TS=%.3f  %s=%e +/- %e' % \
                        (ts, prefactor.name(), prefactor.value(), prefactor.error())
            self._log_value(gammalib.TERSE, name, value)

        # Initialise results
        colnames = []
        values   = {}

        # Set TS value
        colnames.append('TS')
        values['TS'] = ts

        # Set Nevents
        colnames.append('Nevents')
        values['Nevents'] = nevents

        # Set Npred
        colnames.append('Npred')
        values['Npred'] = npred

        # Gather free full fit parameters
        for model in models:
            model_name = model.name()
            for par in model:
                if par.is_free():

                    # Set parameter name
                    name = model_name+"_"+par.name()

                    # Append value
                    colnames.append(name)
                    values[name] = par.value()

                    # Append error
                    name = 'e_'+name
                    colnames.append(name)
                    values[name] = par.error()

        # Bundle together results
        result = {'colnames': colnames, 'values': values}

        # Return
        return result


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Set test source model for this observation
        self.models(modutils.test_source(self.obs().models(), self._srcname))

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Write models into logger
        self._log_models(gammalib.NORMAL, self.obs().models(), 'Input model')

        # Write header
        self._log_header1(gammalib.TERSE, 'Generate TS distribution')

        # Get number of trials
        ntrials = self['ntrials'].integer()

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:
            args        = [(self, '_trial', i) for i in range(ntrials)]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

        # Continue with regular processing
        for i in range(ntrials):

            # If multiprocessing was used then recover results and put them
            # into the log file
            if self._nthreads > 1:
                result = poolresults[i][0]
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

            # ... otherwise make a trial now
            else:
                result = self._trial(i)

            # Write out trial result
            ioutils.write_csv_row(self['outfile'].filename().url(), i,
                                  result['colnames'], result['values'])

        # Return
        return

    def models(self, models):
        """
        Set model

        Parameters
        ----------
        models : `~gammalib.GModels`
            Model container
        """
        # Copy models
        self.obs().models(models)

        # Return
        return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = cstsdist(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Generates a residual map.
#
# Copyright (C) 2014-2020 Michael Mayer
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import ctools
from cscripts import obsutils


# ============== #
# csresmap class #
# ============== #
class csresmap(ctools.csobservation):
    """
    Generates a residual map
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor
        """
        # Initialise application by calling the appropriate class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise class members
        self._resmap       = None
        self._use_maps     = False
        self._skip_binning = False

        # Return
        return


    # Private methods
    def _get_parameters(self):
        """
        Get parameters from parfile and setup the observation
        """
        # Initialise some flags
        self._use_maps     = False
        self._skip_binning = False

        # If there are no observations in the observation container then check
        # if the "inobs" parameter is a counts cube. If the "inobs" parameter
        # represents a FITS file and if this FITS file is a counts cube then
        # set self._skip_binning=True
        if self.obs().size() == 0 and self['inobs'].filename() != 'NONE':
            filename = gammalib.GFilename(self['inobs'].filename())
            if filename.is_fits():
                cta = gammalib.GCTAObservation()
                cta.load(filename)
                if cta.eventtype() == 'CountsCube':
                    self._skip_binning = True

        # If we have a counts cube, then ask whether we also have a model cube.
        # If a model cube name is given then set self._use_maps=True
        if self._skip_binning:
            self._use_maps = self['modcube'].is_valid()

        # If not two maps are given, proceed to set up observation
        if not self._use_maps:

            # Set observation if not done before
            if self.obs().size() == 0:
                self._require_inobs('csresmap.get_parameters()')
                self.obs(self._get_observations())

            # If we have exactly one binned CTA observation then signal that
            # the binning can be skipped
            if self.obs().size()         == 1 and \
               self.obs()[0].classname() == 'GCTAObservation' and \
               self.obs()[0].eventtype() == 'CountsCube':
                self._skip_binning = True

            # Set models if we have none
            if self.obs().models().size() == 0:
                self.obs().models(self['inmodel'].filename())

            # If no binning is provided in the observation then query now
            # the counts cube binning parameters
            if not self._skip_binning:
                self['xref'].real()
                self['yref'].real()
                self['coordsys'].string()
                self['proj'].string()
                self['nxpix'].integer()
                self['nypix'].integer()
                self['binsz'].real()
                self['ebinalg'].string()
                if self['ebinalg'].string() == 'FILE':
                    self['ebinfile'].filename()
                else:
                    self['emin'].real()
                    self['emax'].real()
                    self['enumbins'].integer()
                    if self['ebinalg'].string() == 'POW':
                        self['ebingamma'].real()

        # Query parameters
        self['edisp'].boolean()
        self['algorithm'].string()
        self['publish'].boolean()
        self['chatter'].integer()
        self['clobber'].boolean()
        self['debug'].boolean()

        # Query ahead output parameters
        if (self._read_ahead()):
            self['outmap'].filename()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Return
        return


    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write observation into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Observation')

        # If a counts and model cube are specified then load them as sky map
        if self._use_maps:
            countmap = gammalib.GSkyMap(self['inobs'].filename())
            modelmap = gammalib.GSkyMap(self['modcube'].filename())

        # ... otherwise build a counts cube and model cube
        else:

            # Do not build counts cube if we have already one in the observation
            # container
            if self._skip_binning:
                cta_counts_cube = gammalib.GCTAEventCube(self.obs()[0].events())

            # ... otherwise generate one now from the event list
            else:

                # Write header
                self._log_header1(gammalib.TERSE, 'Generate binned map (ctbin)')

                # Create counts cube
                cta_counts_cube = obsutils.create_counts_cube(self, self.obs())

            # Assign GCTAEventCube to skymap
            countmap = cta_counts_cube.counts()

            # Write header
            self._log_header1(gammalib.TERSE, 'Generate model map (ctmodel)')

            # Create model map
            model = ctools.ctmodel(self.obs())
            model.cube(cta_counts_cube)
            model['chatter'] = self['chatter'].integer()
            model['clobber'] = self['clobber'].boolean()
            model['debug']   = self['debug'].boolean()
            model['edisp']   = self['edisp'].boolean()
            model.run()

            # Get model map into GSkyMap object
            modelmap = model.cube().counts().copy()

        # Calculate residual maps
        # Note that we need a special
        # construct here to avoid memory leaks. This seems to be a SWIG feature
        # as SWIG creates a new object when calling binning.cube()
        countmap1 = countmap.copy()
        countmap1.stack_maps()
        modelmap.stack_maps()
        self._resmap = obsutils.residuals(self,countmap1,modelmap)

        # Optionally publish map
        if self['publish'].boolean():
            self.publish()

        # Return
        return

    def save(self):
        """
        Save residual map
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save residual map')

        # Get outmap parameter
        outmap = self['outmap'].filename()
        
        # Continue only filename and residual map are valid
        if self._resmap != None:

            # Log file name
            self._log_value(gammalib.TERSE, 'Residual map file', outmap.url())

            # Save residual map
            self._resmap.save(outmap, self['clobber'].boolean())

        # Return
        return

    def publish(self, name=''):
        """
        Publish residual map

        Parameters
        ----------
        name : str, optional
            Name of residual map
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Publish residual map')

        # Continue only if residual map is valid
        if self._resmap != None:
        
            # Set default name is user name is empty
            if not name:
                user_name = self._name()
            else:
                user_name = name

            # Log map name
            self._log_value(gammalib.TERSE, 'Residual map name', user_name)

            # Publish map
            self._resmap.publish(user_name)

        # Return
        return

    def models(self, models):
        """
        Set model
        """
        # Copy models
        self.obs().models(models)

        # Return
        return

    def resmap(self):
        """
        Return residual map

        Returns
        -------
        map : `~gammalib.GSkyMap'
            Residual sky map
        """
        # Return
        return self._resmap


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csresmap(sys.argv)

    # Execute application
    app.execute()



#! /usr/bin/env python
# ==========================================================================
# Display Redistribution Matrix File
#
# Copyright (C) 2018-2019 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import sys
import gammalib
import cscripts
try:
    import matplotlib.pyplot as plt
    plt.figure()
    plt.close()
except (ImportError, RuntimeError):
    print('This script needs the "matplotlib" module')
    sys.exit()


# ======== #
# Plot RMF #
# ======== #
def plot_rmf(rmf, plotfile):
    """
    Plot Redistribution Matrix File

    Parameters
    ----------
    rmf : `~gammalib.GRmf`
        Redistribution Matrix File
    plotfile : str
        Plot filename
    """
    # Build title

    # Set energy range
    etrue_min = rmf.etrue().emin().log10TeV()
    etrue_max = rmf.etrue().emax().log10TeV()
    ereco_min = rmf.emeasured().emin().log10TeV()
    ereco_max = rmf.emeasured().emax().log10TeV()
    aspect    = (etrue_max-etrue_min)/(ereco_max-ereco_min)

    # Set number of bins
    netrue = rmf.etrue().size()
    nereco = rmf.emeasured().size()

    # Set image
    image = []
    for ireco in range(nereco-1,-1,-1):
        row = []
        for itrue in range(netrue):
            row.append(rmf[itrue,ireco])
        image.append(row)

    # Create figure
    plt.figure()

    # Plot image
    c    = plt.imshow(image, extent=[etrue_min,etrue_max,ereco_min,ereco_max],
                      interpolation='nearest', aspect=aspect)
    plt.colorbar(c, orientation='vertical', shrink=0.8)

    # Plot title and axis
    plt.title('Redistribution Matrix File')
    plt.xlabel('log10(Etrue/TeV)')
    plt.ylabel('log10(Ereco/TeV)')

    # Show plots or save it into file
    if len(plotfile) > 0:
        plt.savefig(plotfile)
    else:
        plt.show()

    # Return
    return


# ======== #
# Show RMF #
# ======== #
def show_rmf():
    """
    Show Redistribution Matrix File
    """
    # Set usage string
    usage = 'show_rmf.py [-p plotfile] file'

    # Set default options
    options = [{'option': '-p', 'value': ''}]

    # Get arguments and options from command line arguments
    args, options = cscripts.ioutils.get_args_options(options, usage)

    # Extract script parameters from options
    plotfile = options[0]['value']

    # Get RMF
    rmf = gammalib.GRmf(args[0])
    
    # Plot RMF
    plot_rmf(rmf, plotfile)

    # Return
    return


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Show RMF
    show_rmf()



#! /usr/bin/env python
# ==========================================================================
# This scripts performs unit tests for the cstsmapmerge script
#
# Copyright (C) 2016-2018 Juergen Knoedlseder
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import gammalib
import cscripts
from testing import test


# ================================== #
# Test class for cstsmapmerge script #
# ================================== #
class Test(test):
    """
    Test class for cstsmapmerge script

    This test class makes unit tests for the cstsmapmerge script by using it
    from the command line and from Python.
    """
    
    # Constructor
    def __init__(self):
        """
        Constructor
        """
        # Call base class constructor
        test.__init__(self)

        # Set members
        self._map0    = self._datadir + '/tsmap_0.fits'
        self._map1    = self._datadir + '/tsmap_1.fits'
        self._mapstar = self._datadir + '/tsmap_*.fits'
        self._mapat   = '@' + self._datadir + '/tsmaps.txt'

        # Return
        return

    # Set test functions
    def set(self):
        """
        Set all test functions
        """
        # Set test name
        self.name('cstsmapmerge')

        # Append tests
        self.append(self._test_cmd, 'Test cstsmapmerge on command line')
        self.append(self._test_python, 'Test cstsmapmerge from Python')

        # Return
        return

    # Test cstsmapmerge on command line
    def _test_cmd(self):
        """
        Test cstsmapmerge on the command line
        """
        # Set script name
        cstsmapmerge = self._script('cstsmapmerge')

        # Setup cstsmapmerge command
        cmd = cstsmapmerge+' inmaps="'+self._map0+';'+self._map1+'"'+ \
                           ' outmap="cstsmapmerge_cmd1.fits"'+ \
                           ' logfile="cstsmapmerge_cmd1.log" chatter=1'

        # Check if execution was successful
        self.test_assert(self._execute(cmd) == 0,
             'Check successful execution from command line')

        # Check result file
        self._check_result_file('cstsmapmerge_cmd1.fits')

        # Setup cstsmapmerge command
        cmd = cstsmapmerge+' inmaps="map_that_does_not_exist.fits"'+ \
                           ' outmap="cstsmapmerge_cmd2.fits"'+ \
                           ' logfile="cstsmapmerge_cmd2.log" debug=yes chatter=2'

        # Check if execution failed
        self.test_assert(self._execute(cmd, success=False) != 0,
             'Check invalid input file when executed from command line')

        # Check cstsmapmerge --help
        self._check_help(cstsmapmerge)

        # Return
        return

    # Test cstsmapmerge from Python
    def _test_python(self):
        """
        Test cstsmapmerge from Python
        """
        # Set-up semi-colon separated cstsmapmerge
        merge = cscripts.cstsmapmerge()
        merge['inmaps']  = self._map0+';'+self._map1
        merge['outmap']  = 'cstsmapmerge_py1.fits'
        merge['logfile'] = 'cstsmapmerge_py1.log'
        merge['chatter'] = 2

        # Run cstsmapmerge script
        merge.logFileOpen()   # Make sure we get a log file
        merge.run()
        merge.save()

        # Check merged TS map
        self._check_result_file('cstsmapmerge_py1.fits')

        # Set-up space-separated cstsmapmerge
        merge = cscripts.cstsmapmerge()
        merge['inmaps']  = self._map0+' '+self._map1
        merge['outmap']  = 'cstsmapmerge_py2.fits'
        merge['logfile'] = 'cstsmapmerge_py2.log'
        merge['chatter'] = 3

        # Run cstsmapmerge script
        merge.execute()

        # Check merged TS map
        self._check_result_file('cstsmapmerge_py2.fits')

        # Set-up wildcard string cstsmapmerge
        merge = cscripts.cstsmapmerge()
        merge['inmaps']  = self._mapstar
        merge['outmap']  = 'cstsmapmerge_py3.fits'
        merge['logfile'] = 'cstsmapmerge_py3.log'
        merge['chatter'] = 4

        # Run cstsmapmerge script
        merge.execute()

        # Check merged TS map
        self._check_result_file('cstsmapmerge_py3.fits')

        # Set-up ASCII file cstsmapmerge
        merge = cscripts.cstsmapmerge()
        merge['inmaps']  = self._mapat
        merge['outmap']  = 'cstsmapmerge_py4.fits'
        merge['logfile'] = 'cstsmapmerge_py4.log'
        merge['chatter'] = 4

        # Run cstsmapmerge script
        merge.execute()

        # Check merged TS map
        self._check_result_file('cstsmapmerge_py4.fits')
        
        # Set-up cstsmapmerge with incomplete file list
        merge = cscripts.cstsmapmerge()
        merge['inmaps']  = self._map0 + ' ' + self._events
        merge['outmap']  = 'cstsmapmerge_py5.fits'
        merge['logfile'] = 'cstsmapmerge_py5.log'
        merge['chatter'] = 4

        # Run cstsmapmerge script
        merge.execute()

        # Check merged TS map
        self._check_result_file('cstsmapmerge_py5.fits', False)

        # Return
        return

    # Check result file
    def _check_result_file(self, filename, test_complete = True):
        """
        Check result file
        """
        # Open result file
        fits = gammalib.GFits(filename)

        # Get HDUs
        ts        = fits['Primary']
        prefactor = fits['Prefactor']
        index     = fits['Index']
        status    = fits['STATUS MAP']

        # Check dimensions
        self.test_value(ts.naxis(), 2, 'Check for 2 dimensions')
        self.test_value(ts.naxes(0), 2, 'Check for 2 pixels in X')
        self.test_value(ts.naxes(1), 1, 'Check for 1 pixel in Y')
        self.test_value(prefactor.naxis(), 2, 'Check for 2 dimensions')
        self.test_value(prefactor.naxes(0), 2, 'Check for 2 pixels in X')
        self.test_value(prefactor.naxes(1), 1, 'Check for 1 pixel in Y')
        self.test_value(index.naxis(), 2, 'Check for 2 dimensions')
        self.test_value(index.naxes(0), 2, 'Check for 2 pixels in X')
        self.test_value(index.naxes(1), 1, 'Check for 1 pixel in Y')
        self.test_value(status.naxis(), 2, 'Check for 2 dimensions')
        self.test_value(status.naxes(0), 2, 'Check for 2 pixels in X')
        self.test_value(status.naxes(1), 1, 'Check for 1 pixel in Y')
        
        # Initialise flag if map is complete
        done = True
        
        # Loop over status sky map
        for pix in status:
            
            # Check if pix is larger than threshold
            if pix < -0.5:
                done = False
                break
        
        # Test for completeness of merged map
        self.test_assert(done == test_complete, 'Test if map was merged completely')           

        # Return
        return



#! /usr/bin/env python
# ==========================================================================
# Spectral component separation script
#
# Copyright (C) 2020 Luigi Tibaldo
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# ==========================================================================
import ctools
import gammalib
import math
import sys
from cscripts import mputils
from cscripts import obsutils


# =========== #
# csscs class #
# =========== #
class csscs(ctools.csobservation):
    """
    Spectral component separation script
    """

    # Constructor
    def __init__(self, *argv):
        """
        Constructor

        Parameters
        ----------
        argv : list of str
            List of IRAF command line parameter strings of the form
            ``parameter=3``.
        """
        # Initialise application by calling the base class constructor
        self._init_csobservation(self.__class__.__name__, ctools.__version__, argv)

        # Initialise data members
        self._nthreads      = 0
        self._srcnames      = []
        self._method        = None
        self._fits          = None
        self._excl_reg_map  = None  # Exclusion region map for on/off analysis
        self._excl_reg_name = None

        # Return
        return

    # State methods for pickling
    def __getstate__(self):
        """
        Extend ctools.csobservation __getstate__ method

        Returns
        -------
        state : dict
            Pickled instance
        """
        # Set pickled dictionary
        state = {'base':          ctools.csobservation.__getstate__(self),
                 'nthreads':      self._nthreads,
                 'srcnames':      self._srcnames,
                 'method':        self._method,
                 'fits':          self._fits,
                 'excl_reg_map':  self._excl_reg_map,
                 'excl_reg_name': self._excl_reg_name}

        # Return pickled dictionary
        return state

    def __setstate__(self, state):
        """
        Extend ctools.csobservation __setstate__ method

        Parameters
        ----------
        state : dict
            Pickled instance
        """
        # Set state
        ctools.csobservation.__setstate__(self, state['base'])
        self._nthreads      = state['nthreads']
        self._srcnames      = state['srcnames']
        self._method        = state['method']
        self._fits          = state['fits']
        self._excl_reg_map  = state['excl_reg_map']
        self._excl_reg_name = state['excl_reg_name']

        # Return
        return

    # Private methods
    def _get_onoff_parameters(self):
        """
        Get On/Off analysis parameters.

        This is done here rather than in ctools.is_onoff because the exclusion
        map needs to be handled differently as an automatic parameter but
        queried only if not already set and we need to verify is not empty.
        Also many parameters are already queried for all methods

        TODO: verify if this can be made more uniform with other scripts
        """
        # Exclusion map
        if (self._excl_reg_map is not None) and (self._excl_reg_map.map().npix() > 0):
            # Exclusion map set and is not empty
            pass
        elif self['inexclusion'].is_valid():
            self._excl_reg_name = self['inexclusion'].filename()
            self._excl_reg_map = gammalib.GSkyRegionMap(self._excl_reg_name)
        else:
            msg = 'csscs in On/Off mode requires input exclusion region.'
            raise RuntimeError(msg)

        # Other csphagen parameters
        self['enumbins'].integer()
        if self['bkgmethod'].string() == 'REFLECTED':
            self['bkgregmin'].integer()
            self['bkgregskip'].integer()
        self['maxoffset'].real()
        self['etruemin'].real()
        self['etruemax'].real()
        self['etruebins'].integer()

        # Return
        return

    def _get_parameters(self):
        """
        Get parameters from parfile
        """
        # Set observation if not done before
        if self.obs().is_empty():
            self.obs(self._get_observations())

        # Set observation statistic
        self._set_obs_statistic(gammalib.toupper(self['statistic'].string()))

        # Collect number of unbinned, binned and On/Off observations in
        # observation container
        n_unbinned = 0
        n_binned   = 0
        n_onoff    = 0
        for obs in self.obs():
            if obs.classname() == 'GCTAObservation':
                if obs.eventtype() == 'CountsCube':
                    n_binned += 1
                else:
                    n_unbinned += 1
            elif obs.classname() == 'GCTAOnOffObservation':
                n_onoff += 1
        n_cta   = n_unbinned + n_binned + n_onoff
        n_other = self.obs().size() - n_cta

        # Check if there are On/Off or non-IACT observations
        if n_onoff > 0 or n_other > 0:
            msg = 'On/Off or non-CTA observations found. ' \
                  'csscs does not support this type of observations.'
            raise RuntimeError(msg)

        # ... otherwise check if there is a mix of binned and unbinned
        elif n_unbinned > 0 and n_binned > 0:
            msg = 'Mix of unbinned and binned CTA observations ' \
                  'found in observation container. csscs does not ' \
                  'support this mix.'
            raise RuntimeError(msg)

        # Set models if there are none in the container
        if self.obs().models().size() == 0:
            self.obs().models(self['inmodel'].filename())

        # Query the map definition parameters
        self['xref'].real()
        self['yref'].real()
        self['coordsys'].string()
        self['proj'].string()
        self['nxpix'].integer()
        self['nypix'].integer()
        self['binsz'].real()

        # Query radius of ROI for component separation
        self['rad'].real()

        # Query energy boundaries
        self['emin'].real()
        self['emax'].real()

        # If we have unbinned observations query analysis method
        if n_unbinned > 0:
            self._method = self['method'].string()
            # If method is Onoff query csphagen parameters
            if self._method == 'ONOFF':
                self._get_onoff_parameters()
        # Otherwise set method to binned
        else:
            self._method = 'BINNED'

        # Query target source names
        srcnames = self['srcnames'].string()

        # Fashion target sources into Python list. Strip leading and trailing
        # spaces from names.
        self._srcnames = srcnames.split(';')
        for s, name in enumerate(self._srcnames):
            self._srcnames[s] = name.strip()

        # Verify that the target sources are in the model
        src_in_model = True
        for name in self._srcnames:
            src_in_model = src_in_model and self.obs().models().contains(name)
        if src_in_model == False:
            msg = 'Not all target sources are present in input model.'
            raise RuntimeError(msg)

        # For On/Off analysis check the number of gamma-ray sources in the model
        if self._method == 'ONOFF':

            # Initialise number of sources
            nsources = 0

            # Loop over models and verify if model is of type GModelSky
            for model in self.obs().models():
                if model.classname() == 'GModelSky':
                    nsources += 1

            # If there are background gamma-ray sources in the model
            # throw runtime error
            if nsources > len(self._srcnames):
                msg = 'Background gamma-ray sources found in the model. ' \
                      'On/Off analysis does not support this feature.'
                raise RuntimeError(msg)

            # ... otherwise continue
            else:
                pass

        # Query other hidden parameters, just in case
        self['edisp'].boolean()
        self['calc_ulim'].boolean()
        self['calc_ts'].boolean()
        self['fix_bkg'].boolean()
        self['fix_srcs'].boolean()

        # Read ahead output parameters
        if self._read_ahead():
            self['outfile'].query()

        #  Write input parameters into logger
        self._log_parameters(gammalib.TERSE)

        # Set number of processes for multiprocessing
        self._nthreads = mputils.nthreads(self)

        # Return
        return

    def _create_map(self):
        """
        Create a gammalib.GSkyMap object to store the output

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            Sky map
        """
        # Create sky map
        skymap = gammalib.GSkyMap(self['proj'].string(), self['coordsys'].string(),
                                  self['xref'].real(), self['yref'].real(),
                                  -self['binsz'].real(), self['binsz'].real(),
                                  self['nxpix'].integer(), self['nypix'].integer(),
                                  1)

        # Return sky map
        return skymap

    def _adjust_model_pars(self):
        """
        Adjust model parameters
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Adjust model parameters')

        # Adjust model parameters based on input user parameters
        for model in self.obs().models():

            # Initialise TS flag for all models to false
            model.tscalc(False)

            # Log model name
            self._log_header3(gammalib.EXPLICIT, model.name())

            # In On/Off mode we cannot support multiple source morphologies
            # Thus, if we have more than one target we will set them all
            # to isotropic
            if self._method == 'ONOFF' and len(self._srcnames) > 1:

                # Check if it is a source
                try:
                    gammalib.GModelSky(model)
                    # In this case check if the spatial model is already isotropic
                    if model.spatial() == 'DiffuseIsotropic':
                        pass

                    # ... otherwise change it to isotropic
                    else:
                        msg = ' Setting spatial model to diffuse isotropic'
                        self._log_string(gammalib.EXPLICIT, msg)
                        model.spatial(gammalib.GModelSpatialDiffuseConst())
                except:
                    pass

            # Freeze all parameters except the normalization
            # which is assumed to be the first spectral parameter
            normpar = model.spectral()[0]
            for par in model:
                if par.name() == normpar.name():
                    pass
                else:
                    if par.is_free():
                        self._log_string(gammalib.EXPLICIT, ' Fixing "' + par.name() + '"')
                        par.fix()

            # Deal with the target sources
            if model.name() in self._srcnames:

                # Free the normalisation parameter which is assumed to be
                # the first spectral parameter
                if normpar.is_fixed():
                    self._log_string(gammalib.EXPLICIT, ' Freeing "' + normpar.name() + '"')
                    normpar.free()

                # Optionally compute Test Statistic value
                if self['calc_ts'].boolean():
                    model.tscalc(True)

            # Deal with background models
            elif self['fix_bkg'].boolean() and not model.classname() == 'GModelSky':

                if normpar.is_free():
                    self._log_string(gammalib.EXPLICIT, ' Fixing "' + normpar.name() + '"')
                    normpar.fix()


            # Deal with background sources
            elif self['fix_srcs'].boolean() and model.classname() == 'GModelSky':

                if normpar.is_free():
                    self._log_string(gammalib.EXPLICIT, ' Fixing "' + normpar.name() + '"')
                    normpar.fix()

        # Return
        return

    def _mask_cube(self, obs, ra, dec, rad):
        """
        Mask cube observation

        Parameters
        ----------
        obs : `gammalib.GCTAObservation`
            Input observation of type cube
        ra : float
            Right Ascension (deg)
        dec : float
            Declination (deg)
        rad : float
            Radius (deg)

        Returns
        -------
        new_obs : `gammalib.GCTAObservations`
            Output observations with masked cubes
        """
        # Filter cube according to ROI and user energy range
        cubemask = ctools.ctcubemask(obs)
        cubemask['ra'] = ra
        cubemask['dec'] = dec
        cubemask['rad'] = rad
        cubemask['regfile'] = 'NONE'
        cubemask['emin'] = self['emin'].real()
        cubemask['emax'] = self['emax'].real()

        # If chatter level is verbose and debugging is requested then
        # switch also on the debug model in ctcubemask
        if self._logVerbose() and self._logDebug():
            cubemask['debug'] = True

        # Mask cube
        cubemask.run()

        # Get deep copy of filtered observations
        new_obs = cubemask.obs().copy()

        # Return
        return new_obs

    def _mask_evlist(self, obs, ra, dec, rad):
        """
        Mask event list

        Parameters
        ----------
        obs : `gammalib.GCTAObservations`
            Input observations of type event list
        ra : float
            Right Ascension (deg)
        dec : float
            Declination (deg)
        rad : float
            Radius (deg)

        Returns
        -------
        new_obs : `gammalib.GCTAObservations`
            Output observations with masked event lists
        """
        # Filter event list according to ROI and user energy range
        select = ctools.ctselect(obs)
        select['usepnt'] = False
        select['ra']     = ra
        select['dec']    = dec
        select['rad']    = rad
        select['emin']   = self['emin'].real()
        select['emax']   = self['emax'].real()
        select['tmin']   = 'INDEF'
        select['tmax']   = 'INDEF'

        # If chatter level is verbose and debugging is requested then
        # switch also on the debug model in ctcubemask
        if self._logVerbose() and self._logDebug():
            select['debug'] = True

        # Select event list
        select.run()

        # Get deep copy of filtered observation
        new_obs = select.obs().copy()

        # Return
        return new_obs

    def _mask_observations(self, ra, dec, rad):
        """
        Create observations restricted to circular ROI

        Parameters
        ----------
        ra : float
            Right Ascension (deg)
        dec : float
            Declination (deg)
        rad : float
            Radius (deg)

        Returns
        -------
        new_obs : `~gammalib.GObservations`
            Observations in circular ROI
        """
        # Determine type of masking according to observation type and analysis
        # method
        if self._method == 'UNBINNED':
            new_obs = self._mask_evlist(self.obs(), ra, dec, rad)
        elif self._method == 'BINNED':
            new_obs = self._mask_cube(self.obs(), ra, dec, rad)
        elif self._method == 'ONOFF':
            new_obs = obsutils.get_onoff_obs(self, self.obs(), nthreads=1,
                                             ra=ra, dec=dec,
                                             srcname=self._srcnames[0])

        # Return
        return new_obs

    def _pixel_analysis(self, inx):
        """
        Performs analysis over the region of interest corresponding to a
        single pixel of output map

        Parameters
        ----------
        inx : int
            Pixel index

        Returns
        -------
        result : dict
            Results
        """

        # Create template for output map to extract pixel coordinates
        outmap = self._create_map()

        # Determine pixel centre coordinates
        pixdir = outmap.inx2dir(inx)
        ra     = pixdir.ra_deg()
        dec    = pixdir.dec_deg()

        # Write header for spatial bin
        msg = 'Spatial bin (%d) centred on (R.A.,Dec) = (%f,%f) deg' % (inx, ra, dec)
        self._log_header2(gammalib.TERSE, msg)

        # Initialize results
        result = {}
        for name in self._srcnames:
            result[name] = {'flux': 0.0,
                            'flux_err': 0.0,
                            'TS': 0.0,
                            'ulimit': -1.0}

        # Mask observations
        self._log_header3(gammalib.NORMAL, 'Masking observations')
        masked_obs = self._mask_observations(ra, dec, self['rad'].real())

        # Continue only if we have at least one observation
        if masked_obs.size() > 0:

            # Set up likelihood analysis
            self._log_header3(gammalib.NORMAL, 'Performing fit in region')
            like = ctools.ctlike(masked_obs)
            like['edisp']    = self['edisp'].boolean()
            like['nthreads'] = 1  # Avoids OpenMP conflict

            # If chatter level is verbose and debugging is requested then
            # switch also on the debug model in ctlike
            if self._logVerbose() and self._logDebug():
                like['debug'] = True

            # Perform maximum likelihood fit
            like.run()

            # Write model results for explicit chatter level
            self._log_string(gammalib.EXPLICIT, str(like.obs().models()))

            # Continue only if log-likelihood is non-zero
            if like.obs().logL() != 0.0:

                # Prepare objects for flux extraction
                # ROI
                centre = gammalib.GSkyDir()
                centre.radec_deg(ra, dec)
                roi = gammalib.GSkyRegionCircle(centre, self['rad'].real())

                # Energy boundaries
                emin = gammalib.GEnergy(self['emin'].real(), 'TeV')
                emax = gammalib.GEnergy(self['emax'].real(), 'TeV')

                # Loop over target sources
                for name in self._srcnames:

                    # Get source
                    source = like.obs().models()[name]

                    # Get flux
                    # Integrate spectral model between emin and emax
                    flux = source.spectral().flux(emin, emax)

                    # Calculate correction factor
                    # Spatial model flux over ROI divided by ROI solid angle
                    corr_factor  = source.spatial().flux(roi)
                    corr_factor /= gammalib.twopi * (1.0 - math.cos(math.radians(self['rad'].real())))

                    # Multiply flux by correction factor
                    flux                *= corr_factor
                    result[name]['flux'] = flux

                    # Get flux error
                    # Normalization parameter
                    normpar = source.spectral()[0]

                    # Only normalization parameter free
                    # Relative error on flux is same as on normalization
                    # Avoid zero division error
                    if normpar.value() > 0.:
                        flux_error = flux * normpar.error() / normpar.value()
                    else:
                        flux_error = 0.
                    result[name]['flux_error'] = flux_error

                    # If requested get TS
                    if self['calc_ts'].boolean():
                        result[name]['TS'] = source.ts()

                    # If requested compute upper flux limit
                    if self['calc_ulim'].boolean():

                        # Logging information
                        self._log_header3(gammalib.NORMAL,
                                          'Computing upper limit for source ' + name)

                        # Create upper limit object
                        ulimit = ctools.ctulimit(like.obs())
                        ulimit['srcname'] = name
                        ulimit['emin'] = self['emin'].real()
                        ulimit['emax'] = self['emax'].real()

                        # If chatter level is verbose and debugging is requested
                        # then switch also on the debug model in ctulimit
                        if self._logVerbose() and self._logDebug():
                            ulimit['debug'] = True

                        # Try to run upper limit and catch exceptions
                        try:
                            ulimit.run()
                            ulimit_value = ulimit.flux_ulimit()
                            # Multiply by correction factor to get flux per solid angle in ROI
                            ulimit_value *= corr_factor
                            result[name]['ulimit'] = ulimit_value
                        except:
                            self._log_string(gammalib.NORMAL, 'Upper limit '
                                                              'calculation failed.')

            # if likelihood is zero
            else:
                value = 'Likelihood is zero. Bin is skipped.'
                self._log_value(gammalib.TERSE, '(R.A.,Dec) = (%f,%f) deg' % (ra, dec), value)

        # if observation size = 0
        else:
            value = 'Size of masked observations is zero. Bin is skipped.'
            self._log_value(gammalib.TERSE, '(R.A.,Dec) = (%f,%f) deg' % (ra, dec), value)

        # Return result
        return result

    def _write_hdu_keywords(self, hdu, is_flux=False):
        """
        Append cards to HDU

        Parameters
        ----------
        hdu : `~gammalib.GFitsHDU`
            HDU
        """
        # Flux units
        if is_flux:
            hdu.card('BUNIT', 'ph/cm2/s/sr', 'Photon flux')

        # Minimum and maximum energy
        hdu.card('E_MIN', self['emin'].real(), '[TeV] Lower energy boundary')
        hdu.card('E_MAX', self['emax'].real(), '[TeV] Upper energy boundary')
        hdu.card('EUNIT', 'TeV', 'Units for E_MIN and E_MAX')

        # ROI size
        hdu.card('ROISZ', self['rad'].real(),
                 '[deg] Size of ROI used for component separation')

        # Analysis method
        hdu.card('METHOD', self._method, 'Analysis method')

        # For On/Off log information
        if self._method == 'ONOFF':

            # Exclusion map
            if self._excl_reg_name is not None:
                hdu.card('EXCLMAP', self._excl_reg_name.url(), 'Exclusion map name')

            # Model background?
            hdu.card('BKGMODEL', self['use_model_bkg'].boolean(), 'Use background model?')

            # Fit statistic
            hdu.card('STAT', self['statistic'].string(), 'Fit statistic')

        # Return
        return

    def _fill_fits(self, results):
        """
        Fill FITS object to store the results

        Parameters
        ----------
        result : `dict`
            Results

        Returns
        -------
        fits : `~gammalib.GFits`
            FITS object
        """
        # Create new Fits object
        fits = gammalib.GFits()

        # Append empty primary HDU
        fits.append(gammalib.GFitsImageDouble())

        # Loop over target sources
        for s, name in enumerate(self._srcnames):

            # Create skymaps to store results
            # Maps for TS and upper limits created and filled to avoid if statements
            # Will not be saved in the end if not requested
            fmap   = self._create_map()
            errmap = self._create_map()
            tsmap  = self._create_map()
            ulmap  = self._create_map()

            # Loop over pixels and fill maps
            for inx in range(fmap.npix()):
                fmap[inx, 0]   = results[inx][name]['flux']
                errmap[inx, 0] = results[inx][name]['flux_error']
                tsmap[inx, 0]  = results[inx][name]['TS']
                ulmap[inx, 0]  = results[inx][name]['ulimit']

            # Uppercase name to adhere to gammalib convention
            Name = gammalib.toupper(name)

            # Write maps to Fits
            fhdu = fmap.write(fits, Name + ' FLUX')
            self._write_hdu_keywords(fhdu, is_flux=True)
            errhdu = errmap.write(fits, Name + ' FLUX ERROR')
            self._write_hdu_keywords(errhdu, is_flux=True)

            # If requested write TS map to fits
            if self['calc_ts'].boolean():
                tshdu = tsmap.write(fits, Name + ' TS')
                self._write_hdu_keywords(tshdu)

            # If requested write upper limit map to fits
            if self['calc_ulim'].boolean():
                ulhdu = ulmap.write(fits, Name + ' FLUX UPPER LIMIT')
                self._write_hdu_keywords(ulhdu, is_flux=True)

        # Return
        return fits

    def _get_skymap(self, name, quantity):
        """
        Fetches skymap from FITS container

        Parameters
        ----------
        name : str
            Source name
        quantity : str
            Quantity

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            Skymap

        Raises
        ------
        NameError
        """

        # Initialise skymap object
        skymap = None

        # Proceed only if fits has been filled
        if self._fits != None:

            # Check that the name is in the list of target sources
            if name in self._srcnames:

                # Uppercase name
                Name = gammalib.toupper(name)

                # Assemble HDU name
                hduname = Name + ' ' + quantity

                # Try to fetch HDU
                try:
                    hdu = self._fits[hduname]
                    # Convert to GSkyMap
                    skymap = gammalib.GSkyMap(hdu)
                # Catch exception if HDU does not exist
                except:
                    msg = 'HDU "' + hduname + '" not found in FITS container.'
                    raise RuntimeError(msg)

            # ... otherwise throw name error
            else:
                print('ERROR: Source "' + name + '" not found in list of target sources.')
                raise NameError(name)

        # Return
        return skymap

    # Public methods
    def run(self):
        """
        Run the script
        """
        # Switch screen logging on in debug mode
        if self._logDebug():
            self._log.cout(True)

        # Get parameters
        self._get_parameters()

        # Write input observation container into logger
        self._log_observations(gammalib.NORMAL, self.obs(), 'Input observation')

        # Adjust model parameters based on input user parameters
        self._adjust_model_pars()

        # Compute number of pixels of output map
        npix = self._create_map().npix()

        # If more than a single thread is requested then use multiprocessing
        if self._nthreads > 1:

            # Compute values in pixels
            args = [(self, '_pixel_analysis', i)
                    for i in range(npix)]
            poolresults = mputils.process(self._nthreads, mputils.mpfunc, args)

            # Construct results
            results = []
            for i in range(npix):
                results.append(poolresults[i][0])
                self._log_string(gammalib.TERSE, poolresults[i][1]['log'], False)

        # Otherwise loop over pixels and run the pixel analysis
        else:
            results = []
            for i in range(npix):
                results.append(self._pixel_analysis(i))

        # Fill output Fits
        self._fits = self._fill_fits(results)

        # Return
        return

    def save(self):
        """
        Save maps to Fits flie
        """
        # Write header
        self._log_header1(gammalib.TERSE, 'Save maps')

        # Continue only if FITS file is valid
        if self._fits != None:
            # Get outmap parameter
            outfile = self['outfile'].filename()

            # Log file name
            self._log_value(gammalib.NORMAL, 'Output file', outfile.url())

            # Save spectrum
            self._fits.saveto(outfile, self['clobber'].boolean())

        # Return
        return

    def exclusion_map(self, object=None):
        """
        Return and optionally set the exclusion regions map

        Parameters
        ----------
        object : `~gammalib.GSkyRegion` or `~gammalib.GSkyMap` or `~gammalib.GFilename`
            Exclusion regions object

        Returns
        -------
        region : `~gammalib.GSkyRegionMap`
            Exclusion regions map
        """
        # If a regions object is provided then set the regions
        if object is not None:
            self._excl_reg_map = gammalib.GSkyRegionMap(object)

        # Return
        return self._excl_reg_map

    def fits(self):
        """
        Return copy of the fits container

        Returns
        -------
        fits : `~gammalib.GFits`
            FITS file containing the maps
        """
        # Return
        return self._fits.copy()

    def flux(self, name):
        """
        Return flux skymap

        Parameters
        ----------
        name : str
            Source name

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            Flux skymap
        """
        # Get flux skymap
        skymap = self._get_skymap(name, 'FLUX')

        # Return flux sky map
        return skymap

    def flux_error(self, name):
        """
        Return flux error skymap

        Parameters
        ----------
        name : str
            Source name

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            Flux error skymap
        """
        # Get flux error sky map
        skymap = self._get_skymap(name, 'FLUX ERROR')

        # Return flux error sky map
        return skymap

    def ts(self, name):
        """
        Return TS skymap

        Parameters
        ----------
        name : str
            Source name

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            TS skymap
        """
        # Check that TS computation is requested
        if self['calc_ts'].boolean():
            skymap = self._get_skymap(name, 'TS')

        # ... otherwise raise error
        else:
            msg = 'TS computation not requested. Change user parameter ' \
                  '"calc_ts" to True to obtain TS.'
            raise RuntimeError(msg)

        # Return TS map
        return skymap

    def ulimit(self, name):
        """
        Return flux upper limit skymap

        Parameters
        ----------
        name : str
            Source name

        Returns
        -------
        skymap : `~gammalib.GSkyMap`
            Flux upper limit skymap
        """
        # Check that upper limit computation is requested
        if self['calc_ulim'].boolean():
            skymap = self._get_skymap(name, 'FLUX UPPER LIMIT')

        # ... otherwise raise error
        else:
            msg = 'Upper limit computation not requested. Change user parameter ' \
                  '"calc_ulimit" to True to obtain upper limit.'
            raise RuntimeError(msg)

        # Return upper limit map
        return skymap


# ======================== #
# Main routine entry point #
# ======================== #
if __name__ == '__main__':

    # Create instance of application
    app = csscs(sys.argv)

    # Execute application
    app.execute()
